"""Provides an interface for splitting up a large regridding task."""

import esmpy
import numpy as np

from esmf_regrid.constants import Constants
from esmf_regrid.experimental._partial import PartialRegridder
from esmf_regrid.experimental.io import load_regridder, save_regridder
from esmf_regrid.schemes import _get_grid_dims


def _get_chunk(cube, sl):
    if cube.mesh is None:
        grid_dims = _get_grid_dims(cube)
    else:
        grid_dims = (cube.mesh_dim(),)
    full_slice = [np.s_[:]] * len(cube.shape)
    for s, d in zip(sl, grid_dims, strict=True):
        full_slice[d] = np.s_[s[0] : s[1]]
    return cube[*full_slice]


def _determine_blocks(shape, chunks, num_chunks, explicit_blocks):
    which_inputs = (
        chunks is not None,
        num_chunks is not None,
        explicit_blocks is not None,
    )
    if sum(which_inputs) == 0:
        msg = "Partition blocks must must be specified by either chunks, num_chunks, or explicit_chunks."
        raise ValueError(msg)
    if sum(which_inputs) > 1:
        msg = "Potentially conflicting partition block definitions."
        raise ValueError(msg)
    if num_chunks is not None:
        chunks = [s // n for s, n in zip(shape, num_chunks, strict=True)]
        for chunk in chunks:
            if chunk == 0:
                msg = "`num_chunks` cannot divide a dimension into more blocks than the size of that dimension."
                raise ValueError(msg)
    if chunks is not None:
        if all(isinstance(x, int) for x in chunks):
            proper_chunks = []
            for s, c in zip(shape, chunks, strict=True):
                proper_chunk = [c] * (s // c)
                if s % c != 0:
                    proper_chunk += [s % c]
                proper_chunks.append(proper_chunk)
            chunks = proper_chunks
        for s, chunk in zip(shape, chunks, strict=True):
            if sum(chunk) != s:
                msg = "Chunks must sum to the size of their respective dimension."
                raise ValueError(msg)
        bounds = [np.cumsum([0, *chunk]) for chunk in chunks]
        if len(bounds) == 1:
            msg = "Chunks must have exactly two dimensions."
            raise ValueError(msg)
            # TODO: This is currently blocked by the fact that slicing an Iris cube on its mesh dimension
            #  does not currently yield another cube with a mesh. When this is fixed, the following
            #  code can be uncommented and the noqa on the following line can be removed.
            # explicit_blocks = [
            #     [[int(lower), int(upper)]]
            #     for lower, upper in zip(bounds[0][:-1], bounds[0][1:], strict=True)
            # ]
        elif len(bounds) == 2:  # noqa: RET506
            explicit_blocks = [
                [[int(ly), int(uy)], [int(lx), int(ux)]]
                for ly, uy in zip(bounds[0][:-1], bounds[0][1:], strict=True)
                for lx, ux in zip(bounds[1][:-1], bounds[1][1:], strict=True)
            ]
        else:
            msg = "Chunks must not exceed two dimensions."
            raise ValueError(msg)
    if len(explicit_blocks[0]) != len(shape):
        msg = "Dimensionality of blocks does not match the number of dimensions."
        raise ValueError(msg)
    return explicit_blocks


class Partition:
    """Class for breaking down regridding into manageable chunks."""

    def __init__(
        self,
        src,
        tgt,
        scheme,
        file_names,
        use_dask_src_chunks=False,
        src_chunks=None,
        num_src_chunks=None,
        explicit_src_blocks=None,
        auto_generate=False,
        saved_files=None,
    ):
        """Class for breaking down regridding into manageable chunks.

        Parameters
        ----------
        src : cube
            Source cube.
        tgt : cube
            Target cube.
        scheme : regridding scheme
            Regridding scheme to generate regridders, either ESMFAreaWeighted or ESMFBilinear.
        file_names : iterable of str
            A list of file names to save/load parts of the regridder to/from.
        use_dask_src_chunks : bool, default=False
            If true, partition using the same chunks from the source cube.
        src_chunks : numpy array, tuple of int or tuple of tuple of int, default=None
            Specify the size of blocks to use to divide up the cube. Dimensions are specified
            in y,x axis order. If `src_chunks` is a tuple of int, each integer describes
            the maximum size of a block in that dimension. If `src_chunks` is a tuple of tuples,
            each sub-tuple describes the size of each successive block in that dimension. The sum
            of these block sizes in each of the sub-tuples should add up to the total size of that
            dimension or else an error is raised.
        num_src_chunks : tuple of int
            Specify the number of blocks to use to divide up the cube. Dimensions are specified
            in y,x axis order. Each integer describes the number of blocks that dimension will
            be divided into.
        explicit_src_blocks : arraylike NxMx2
            Explicitly specify the bounds of each block in the partition. Describes N blocks
            along M dimensions with a pair of upper and lower bounds. The upper and lower bounds
            describe a slice of an array, e.g. the bounds (3, 6) describe the indices 3, 4, 5 in
            a particular dimension.
        auto_generate : bool, default=False
            When true, start generating files on initialisation.
        saved_files : iterable of str
            A list of paths to previously saved files.

        Warnings
        --------
        This class is still experimental. While we aim to maintain backwards compatibility where
        possible, there is no guarantee that the structure of any generated files will remain
        consistent and compatible with future versions.

        Note
        ----
        The source is partitioned into blocks using one of the four mutually exclusive arguments,
        `use_dask_src_chunks`, `src_chunks`, `num_src_chunks`, or `explicit_src_blocks`. These
        describe a partition into a number of blocks which must equal the number of `file_names`.

        Currently, it is only possible to divide the source grid into chunks.
        Meshes are not yet supported as a source.
        """
        if scheme._method == Constants.Method.NEAREST:
            msg = "The `Nearest` method is not implemented."
            raise NotImplementedError(msg)
        if scheme._method == Constants.Method.BILINEAR:
            pole_method = scheme.esmf_args.get("pole_method")
            if pole_method != esmpy.PoleMethod.NONE:
                msg = (
                    "Bilinear regridding must have a `pole_method` of `esmpy.PoleMethod.NONE` in "
                    "the `esmf_args` in order for Partition to work.`"
                )
                raise ValueError(msg)
        # TODO: Extract a slice of the cube.
        self.src = src
        if src.mesh is None:
            grid_dims = _get_grid_dims(src)
        else:
            msg = "Partition does not yet support source meshes."
            raise NotImplementedError(msg)
            # TODO: This is currently blocked by the fact that slicing an Iris cube on its mesh dimension
            #  does not currently yield another cube with a mesh. When this is fixed, the following
            #  code can be uncommented.
            # grid_dims = (src.mesh_dim(),)
        shape = tuple(src.shape[i] for i in grid_dims)
        self.tgt = tgt
        self.scheme = scheme
        # TODO: consider abstracting away the idea of files
        self.file_names = file_names
        if use_dask_src_chunks:
            if src_chunks is not None:
                msg = "`src_chunks` and `use_dask_src_chunks` cannot be used at the same time."
                raise ValueError(msg)
            if not src.has_lazy_data():
                msg = "If `use_dask_src_chunks=True`, the source cube must be lazy."
                raise TypeError(msg)
            src_chunks = src.slices(grid_dims).next().lazy_data().chunks
        self.src_blocks = _determine_blocks(
            shape, src_chunks, num_src_chunks, explicit_src_blocks
        )
        if len(self.src_blocks) != len(file_names):
            msg = "Number of source blocks does not match number of file names."
            raise ValueError(msg)
        # This will be controllable in future
        tgt_blocks = None
        self.tgt_blocks = tgt_blocks
        if tgt_blocks is not None:
            msg = "Target chunking not yet implemented."
            raise NotImplementedError(msg)

        # Note: this may need to become more sophisticated when both src and tgt are large
        self.file_block_dict = dict(zip(self.file_names, self.src_blocks, strict=True))

        if saved_files is None:
            self.saved_files = []
        else:
            self.saved_files = saved_files
        if auto_generate:
            self.generate_files(self.file_names)

    def __repr__(self):
        """Return a representation of the class."""
        result = (
            f"Partition("
            f"src={self.src!r}, "
            f"tgt={self.tgt!r}, "
            f"scheme={self.scheme}, "
            f"num file_names={len(self.file_names)},"
            f"num saved_files={len(self.saved_files)})"
        )
        return result

    @property
    def unsaved_files(self):
        """List of files not yet generated."""
        return [file for file in self.file_names if file not in self.saved_files]

    def generate_files(self, files_to_generate=None):
        """Generate files with regridding information.

        Parameters
        ----------
        files_to_generate : int, default=None
            Specify the number of files to generate, default behaviour is to generate all files.
        """
        if files_to_generate is None:
            files = self.unsaved_files
        else:
            if not isinstance(files_to_generate, int):
                msg = "`files_to_generate` must be an integer."
                raise ValueError(msg)
            files = self.unsaved_files[:files_to_generate]

        for file in files:
            src_block = self.file_block_dict[file]
            src = _get_chunk(self.src, src_block)
            tgt = self.tgt
            regridder = self.scheme.regridder(src, tgt)
            weights = regridder.regridder.weight_matrix
            regridder = PartialRegridder(
                src, tgt, src_block, None, weights, self.scheme
            )
            save_regridder(regridder, file, allow_partial=True)
            self.saved_files.append(file)

    def apply_regridders(self, cube, allow_incomplete=False):
        """Apply the saved regridders to a cube.

        Parameters
        ----------
        allow_incomplete : bool, default=False
            If False, raise an error if not all files have been generated. If True, perform
            regridding using the files which have been generated.
        """
        # for each target chunk, iterate through each associated regridder
        # for now, assume one target chunk
        if len(self.saved_files) == 0:
            msg = "No files have been generated."
            raise OSError(msg)
        if not allow_incomplete and len(self.unsaved_files) != 0:
            msg = "Not all files have been generated."
            raise OSError(msg)
        current_result = None
        current_weights = None
        files = self.saved_files

        for file, chunk in zip(self.file_names, self.src_blocks, strict=True):
            if file in files:
                next_regridder = load_regridder(file, allow_partial=True)
                cube_chunk = _get_chunk(cube, chunk)
                next_weights, next_result, extra = next_regridder.partial_regrid(
                    cube_chunk
                )
                if current_weights is None:
                    current_weights = next_weights
                else:
                    current_weights += next_weights
                if current_result is None:
                    current_result = next_result
                else:
                    current_result += next_result

        # NOTE: the final "finish_regridding" operation could be performed using any one
        #  of the partial regridders,but the correct "corresponding" slice of the source
        #  must be passed.
        #  See :meth:`~esmf_regrid.experimental._partial.PartialRegridder.finish_regridding`.
        return next_regridder.finish_regridding(
            cube_chunk,  # matches *this* partial regridder
            current_weights,
            current_result,
            extra,  # should be *the same* for all the partial results
        )



"""Unit tests for :mod:`esmf_regrid.experimental.partition`."""

import dask.array as da
import esmpy
import numpy as np
import pytest

from esmf_regrid import ESMFAreaWeighted, ESMFBilinear, ESMFNearest
from esmf_regrid.experimental.partition import Partition
from esmf_regrid.tests.unit.schemes.test__cube_to_GridInfo import (
    _curvilinear_cube,
    _grid_cube,
)
from esmf_regrid.tests.unit.schemes.test__mesh_to_MeshInfo import (
    _gridlike_mesh_cube,
)
from esmf_regrid.tests.unit.schemes.test_regrid_rectilinear_to_rectilinear import (
    _make_full_cubes,
)


def test_Partition(tmp_path):
    """Test basic implementation of Partition class."""
    src = _grid_cube(150, 500, (-180, 180), (-90, 90), circular=True)
    src.data = np.arange(150 * 500).reshape([500, 150])
    tgt = _grid_cube(16, 36, (-180, 180), (-90, 90), circular=True)

    files = [tmp_path / f"partial_{x}.nc" for x in range(5)]
    scheme = ESMFAreaWeighted(mdtol=1)

    blocks = [
        [[0, 100], [0, 150]],
        [[100, 200], [0, 150]],
        [[200, 300], [0, 150]],
        [[300, 400], [0, 150]],
        [[400, 500], [0, 150]],
    ]
    partition = Partition(src, tgt, scheme, files, explicit_src_blocks=blocks)

    partition.generate_files()

    result = partition.apply_regridders(src)
    expected = src.regrid(tgt, scheme)
    assert np.allclose(result.data, expected.data)
    assert result == expected


def test_Partition_block_api(tmp_path):
    """Test API for controlling block shape for Partition class."""
    src = _grid_cube(150, 500, (-180, 180), (-90, 90), circular=True)
    tgt = _grid_cube(16, 36, (-180, 180), (-90, 90), circular=True)

    files = [tmp_path / f"partial_{x}.nc" for x in range(5)]
    scheme = ESMFAreaWeighted(mdtol=1)

    num_src_chunks = (5, 1)
    partition = Partition(src, tgt, scheme, files, num_src_chunks=num_src_chunks)

    expected_chunks = [
        [[0, 100], [0, 150]],
        [[100, 200], [0, 150]],
        [[200, 300], [0, 150]],
        [[300, 400], [0, 150]],
        [[400, 500], [0, 150]],
    ]
    assert partition.src_blocks == expected_chunks

    src_chunks = (100, 150)
    partition = Partition(src, tgt, scheme, files, src_chunks=src_chunks)

    expected_chunks = [
        [[0, 100], [0, 150]],
        [[100, 200], [0, 150]],
        [[200, 300], [0, 150]],
        [[300, 400], [0, 150]],
        [[400, 500], [0, 150]],
    ]
    assert partition.src_blocks == expected_chunks

    src_chunks = ((100, 100, 100, 100, 100), (150,))
    partition = Partition(src, tgt, scheme, files, src_chunks=src_chunks)

    expected_chunks = [
        [[0, 100], [0, 150]],
        [[100, 200], [0, 150]],
        [[200, 300], [0, 150]],
        [[300, 400], [0, 150]],
        [[400, 500], [0, 150]],
    ]
    assert partition.src_blocks == expected_chunks

    src.data = da.from_array(src.data, chunks=src_chunks)
    partition = Partition(src, tgt, scheme, files, use_dask_src_chunks=True)

    expected_chunks = [
        [[0, 100], [0, 150]],
        [[100, 200], [0, 150]],
        [[200, 300], [0, 150]],
        [[300, 400], [0, 150]],
        [[400, 500], [0, 150]],
    ]
    assert partition.src_blocks == expected_chunks


def test_Partition_mesh_src(tmp_path):
    """Test Partition class when the source has a mesh."""
    src = _gridlike_mesh_cube(150, 500)
    src.data = np.arange(150 * 500)
    tgt = _grid_cube(16, 36, (-180, 180), (-90, 90), circular=True)

    files = [tmp_path / f"partial_{x}.nc" for x in range(5)]
    scheme = ESMFAreaWeighted(mdtol=1)

    src_chunks = (15000,)
    with pytest.raises(NotImplementedError):
        _ = Partition(src, tgt, scheme, files, src_chunks=src_chunks)

    # TODO: when mesh partitioning becomes possible, uncomment.
    # expected_src_chunks = [[[0, 15000]], [[15000, 30000]], [[30000, 45000]], [[45000, 60000]], [[60000, 75000]]]
    # assert partition.src_blocks == expected_src_chunks
    #
    # partition.generate_files()
    #
    # result = partition.apply_regridders(src)
    # expected = src.regrid(tgt, scheme)
    # assert np.allclose(result.data, expected.data)
    # assert result == expected


def test_Partition_curv_src(tmp_path):
    """Test Partition class when the source has a curvilinear grid."""
    src = _curvilinear_cube(150, 500, (-180, 180), (-90, 90))
    src.data = np.arange(150 * 500).reshape([500, 150])
    tgt = _grid_cube(16, 36, (-180, 180), (-90, 90), circular=True)

    files = [tmp_path / f"partial_{x}.nc" for x in range(5)]
    scheme = ESMFAreaWeighted(mdtol=1)

    src_chunks = (100, 150)
    partition = Partition(src, tgt, scheme, files, src_chunks=src_chunks)

    expected_src_chunks = [
        [[0, 100], [0, 150]],
        [[100, 200], [0, 150]],
        [[200, 300], [0, 150]],
        [[300, 400], [0, 150]],
        [[400, 500], [0, 150]],
    ]
    assert partition.src_blocks == expected_src_chunks

    partition.generate_files()

    result = partition.apply_regridders(src)
    expected = src.regrid(tgt, scheme)
    assert np.allclose(result.data, expected.data)
    assert result == expected


def test_Partition_bilinear(tmp_path):
    """Test Partition class for bilinear regridding."""
    src = _grid_cube(150, 500, (-180, 180), (-90, 90), circular=True)
    src.data = np.arange(150 * 500).reshape([500, 150])
    tgt = _grid_cube(16, 36, (-180, 180), (-90, 90), circular=True)

    files = [tmp_path / f"partial_{x}.nc" for x in range(5)]
    src_chunks = (100, 150)

    bad_scheme = ESMFBilinear()
    with pytest.raises(ValueError):
        _ = Partition(src, tgt, bad_scheme, files, src_chunks=src_chunks)

    # The pole_method must be NONE for bilinear regridding partitions to work.
    scheme = ESMFBilinear(esmf_args={"pole_method": esmpy.PoleMethod.NONE})

    partition = Partition(src, tgt, scheme, files, src_chunks=src_chunks)

    partition.generate_files()

    result = partition.apply_regridders(src)
    expected = src.regrid(tgt, scheme)
    assert np.allclose(result.data, expected.data)
    assert result == expected


def test_Partition_mesh_tgt(tmp_path):
    """Test Partition class when the target has a mesh."""
    src = _grid_cube(150, 500, (-180, 180), (-90, 90), circular=True)
    src.data = np.arange(150 * 500).reshape([500, 150])
    tgt = _gridlike_mesh_cube(16, 36)

    files = [tmp_path / f"partial_{x}.nc" for x in range(5)]
    scheme = ESMFAreaWeighted(mdtol=1)

    src_chunks = (100, 150)
    partition = Partition(src, tgt, scheme, files, src_chunks=src_chunks)

    partition.generate_files()

    result = partition.apply_regridders(src)
    expected = src.regrid(tgt, scheme)
    assert np.allclose(result.data, expected.data)
    assert result == expected


def test_conflicting_chunks(tmp_path):
    """Test error handling of Partition class."""
    src = _grid_cube(150, 500, (-180, 180), (-90, 90), circular=True)
    tgt = _grid_cube(16, 36, (-180, 180), (-90, 90), circular=True)

    files = [tmp_path / f"partial_{x}.nc" for x in range(5)]
    scheme = ESMFAreaWeighted(mdtol=1)
    num_src_chunks = (5, 1)
    src_chunks = (100, 150)
    blocks = [
        [[0, 100], [0, 150]],
        [[100, 200], [0, 150]],
        [[200, 300], [0, 150]],
        [[300, 400], [0, 150]],
        [[400, 500], [0, 150]],
    ]

    with pytest.raises(ValueError):
        _ = Partition(
            src,
            tgt,
            scheme,
            files,
            src_chunks=src_chunks,
            num_src_chunks=num_src_chunks,
        )
    with pytest.raises(ValueError):
        _ = Partition(
            src, tgt, scheme, files, src_chunks=src_chunks, explicit_src_blocks=blocks
        )
    with pytest.raises(ValueError):
        _ = Partition(src, tgt, scheme, files)
    with pytest.raises(TypeError):
        _ = Partition(src, tgt, scheme, files, use_dask_src_chunks=True)
    with pytest.raises(ValueError):
        _ = Partition(src, tgt, scheme, files[:-1], src_chunks=src_chunks)


def test_multidimensional_cube(tmp_path):
    """Test Partition class when the source has a multidimensional cube."""
    src_cube, tgt_grid, expected_cube = _make_full_cubes()
    files = [tmp_path / f"partial_{x}.nc" for x in range(4)]
    scheme = ESMFAreaWeighted(mdtol=1)
    chunks = (2, 3)

    partition = Partition(src_cube, tgt_grid, scheme, files, src_chunks=chunks)

    partition.generate_files()

    result = partition.apply_regridders(src_cube)

    # Lenient check for data.
    assert np.allclose(result.data, expected_cube.data)

    # Check metadata and coords.
    result.data = expected_cube.data
    assert result == expected_cube


def test_save_incomplete(tmp_path):
    """Test Partition class when a limited number of files are saved."""
    src = _grid_cube(150, 500, (-180, 180), (-90, 90), circular=True)
    tgt = _grid_cube(16, 36, (-180, 180), (-90, 90), circular=True)

    files = [tmp_path / f"partial_{x}.nc" for x in range(5)]
    src_chunks = (100, 150)
    scheme = ESMFAreaWeighted(mdtol=1)
    num_initial_chunks = 3
    expected_files = files[:num_initial_chunks]

    partition = Partition(src, tgt, scheme, files, src_chunks=src_chunks)
    with pytest.raises(OSError):
        _ = partition.apply_regridders(src, allow_incomplete=True)

    partition.generate_files(files_to_generate=num_initial_chunks)
    assert partition.saved_files == expected_files

    expected_array_partial = np.ma.zeros([36, 16])
    expected_array_partial[22:] = np.ma.masked

    with pytest.raises(OSError):
        _ = partition.apply_regridders(src)
    partial_result = partition.apply_regridders(src, allow_incomplete=True)
    assert np.ma.allclose(partial_result.data, expected_array_partial)

    loaded_partition = Partition(
        src, tgt, scheme, files, src_chunks=src_chunks, saved_files=expected_files
    )

    with pytest.raises(OSError):
        _ = loaded_partition.apply_regridders(src)
    partial_result_2 = partition.apply_regridders(src, allow_incomplete=True)
    assert np.ma.allclose(partial_result_2.data, expected_array_partial)

    loaded_partition.generate_files()

    result = loaded_partition.apply_regridders(src)
    expected_array = np.ma.zeros([36, 16])
    assert np.ma.allclose(result.data, expected_array)


def test_nearest_invalid(tmp_path):
    """Test Partition class when initialised with an invalid scheme."""
    src_cube, tgt_grid, _ = _make_full_cubes()
    files = [tmp_path / f"partial_{x}.nc" for x in range(4)]
    scheme = ESMFNearest()
    chunks = (2, 3)

    with pytest.raises(NotImplementedError):
        _ = Partition(src_cube, tgt_grid, scheme, files, src_chunks=chunks)


def test_Partition_repr(tmp_path):
    """Test repr of Partition instance."""
    src_cube, tgt_grid, _ = _make_full_cubes()
    files = [tmp_path / f"partial_{x}.nc" for x in range(4)]
    scheme = ESMFAreaWeighted()
    chunks = (2, 3)

    partition = Partition(src_cube, tgt_grid, scheme, files, src_chunks=chunks)

    expected_repr = (
        "Partition(src=<iris 'Cube' of air_temperature / (K) "
        "(height: 2; latitude: 3; time: 4; longitude: 5; -- : 6)>, "
        "tgt=<iris 'Cube' of unknown / (unknown) (latitude: 5; longitude: 3)>, "
        "scheme=ESMFAreaWeighted(mdtol=0, use_src_mask=False, use_tgt_mask=False, esmf_args={}), "
        "num file_names=4,num saved_files=0)"
    )
    assert repr(partition) == expected_repr



"""Provides load/save functions for regridders."""

from contextlib import contextmanager

import iris
from iris.coords import AuxCoord
from iris.cube import Cube, CubeList
import numpy as np
import scipy.sparse

import esmf_regrid
from esmf_regrid import Constants, _load_context, check_method, esmpy
from esmf_regrid.experimental._partial import PartialRegridder
from esmf_regrid.experimental.unstructured_scheme import (
    GridToMeshESMFRegridder,
    MeshToGridESMFRegridder,
)
from esmf_regrid.schemes import (
    ESMFAreaWeighted,
    ESMFAreaWeightedRegridder,
    ESMFBilinear,
    ESMFBilinearRegridder,
    ESMFNearestRegridder,
    GridRecord,
    MeshRecord,
)

SUPPORTED_REGRIDDERS = [
    ESMFAreaWeightedRegridder,
    ESMFBilinearRegridder,
    ESMFNearestRegridder,
    GridToMeshESMFRegridder,
    MeshToGridESMFRegridder,
    PartialRegridder,
]
_REGRIDDER_NAME_MAP = {rg_class.__name__: rg_class for rg_class in SUPPORTED_REGRIDDERS}
_SOURCE_NAME = "regridder_source_field"
_SOURCE_MASK_NAME = "regridder_source_mask"
_TARGET_NAME = "regridder_target_field"
_TARGET_MASK_NAME = "regridder_target_mask"
_WEIGHTS_NAME = "regridder_weights"
_WEIGHTS_SHAPE_NAME = "weights_shape"
_WEIGHTS_ROW_NAME = "weight_matrix_rows"
_WEIGHTS_COL_NAME = "weight_matrix_columns"
_REGRIDDER_TYPE = "regridder_type"
_VERSION_ESMF = "ESMF_version"
_VERSION_INITIAL = "esmf_regrid_version_on_initialise"
_MDTOL = "mdtol"
_METHOD = "method"
_RESOLUTION = "resolution"
_SOURCE_RESOLUTION = "src_resolution"
_TARGET_RESOLUTION = "tgt_resolution"
_ESMF_ARGS = "esmf_args"
_SRC_SLICE_NAME = "src_slice"
_TGT_SLICE_NAME = "tgt_slice"
_VALID_ESMF_KWARGS = [
    "pole_method",
    "regrid_pole_npoints",
    "line_type",
    "extrap_method",
    "extrap_num_src_pnts",
    "extrap_dist_exponent",
    "extrap_num_levels",
    "unmapped_action",
    "ignore_degenerate",
    "large_file",
]
_POLE_METHOD_DICT = {e.name: e for e in esmpy.PoleMethod}
_LINE_TYPE_DICT = {e.name: e for e in esmpy.LineType}
_EXTRAP_METHOD_DICT = {e.name: e for e in esmpy.ExtrapMethod}
_UNMAPPED_ACTION_DICT = {e.name: e for e in esmpy.UnmappedAction}
_ESMF_ENUM_ARGS = {
    "pole_method": _POLE_METHOD_DICT,
    "line_type": _LINE_TYPE_DICT,
    "extrap_method": _EXTRAP_METHOD_DICT,
    "unmapped_action": _UNMAPPED_ACTION_DICT,
}
_ESMF_BOOL_ARGS = ["ignore_degenerate", "large_file"]


def _add_mask_to_cube(mask, cube, name):
    if isinstance(mask, np.ndarray):
        mask = mask.astype(int)
        mask_coord = AuxCoord(mask, var_name=name, long_name=name)
        cube.add_aux_coord(mask_coord, list(range(cube.ndim)))


@contextmanager
def _managed_var_name(src_cube, tgt_cube):
    src_coord_names = []
    src_mesh_coords = []
    if src_cube.mesh is not None:
        src_mesh = src_cube.mesh
        src_mesh_coords = src_mesh.coords()
        src_coord_names = [coord.var_name for coord in src_mesh_coords]
    tgt_coord_names = []
    tgt_mesh_coords = []
    if tgt_cube.mesh is not None:
        tgt_mesh = tgt_cube.mesh
        tgt_mesh_coords = tgt_mesh.coords()
        tgt_coord_names = [coord.var_name for coord in tgt_mesh_coords]

    try:
        for coord in src_mesh_coords:
            coord.var_name = f"{_SOURCE_NAME}_mesh_{coord.name()}"
        for coord in tgt_mesh_coords:
            coord.var_name = f"{_TARGET_NAME}_mesh_{coord.name()}"
        yield None
    finally:
        for coord, var_name in zip(src_mesh_coords, src_coord_names, strict=False):
            coord.var_name = var_name
        for coord, var_name in zip(tgt_mesh_coords, tgt_coord_names, strict=False):
            coord.var_name = var_name


def _clean_var_names(cube):
    cube.var_name = None
    for coord in cube.coords():
        coord.var_name = None
    if cube.mesh is not None:
        cube.mesh.var_name = None
        for coord in cube.mesh.coords():
            coord.var_name = None
        for con in cube.mesh.connectivities():
            con.var_name = None


def _standard_grid_cube(grid, name):
    if grid[0].ndim == 1:
        shape = [coord.points.size for coord in grid]
    else:
        shape = grid[0].shape
    data = np.zeros(shape)
    cube = Cube(data, var_name=name, long_name=name)
    if grid[0].ndim == 1:
        cube.add_dim_coord(grid[0], 0)
        cube.add_dim_coord(grid[1], 1)
    else:
        cube.add_aux_coord(grid[0], [0, 1])
        cube.add_aux_coord(grid[1], [0, 1])
    return cube


def _standard_mesh_cube(mesh, location, name):
    mesh_coords = mesh.to_MeshCoords(location)
    data = np.zeros(mesh_coords[0].points.shape[0])
    cube = Cube(data, var_name=name, long_name=name)
    for coord in mesh_coords:
        cube.add_aux_coord(coord, 0)
    return cube


def _generate_src_tgt(regridder_type, rg, allow_partial):
    if regridder_type in [
        "ESMFAreaWeightedRegridder",
        "ESMFBilinearRegridder",
        "ESMFNearestRegridder",
        "PartialRegridder",
    ]:
        if regridder_type == "PartialRegridder" and not allow_partial:
            e_msg = "To save a PartialRegridder, `allow_partial=True` must be set."
            raise ValueError(e_msg)
        src_grid = rg._src
        if isinstance(src_grid, GridRecord):
            src_cube = _standard_grid_cube(
                (src_grid.grid_y, src_grid.grid_x), _SOURCE_NAME
            )
        elif isinstance(src_grid, MeshRecord):
            src_mesh, src_location = src_grid
            src_cube = _standard_mesh_cube(src_mesh, src_location, _SOURCE_NAME)
        else:
            e_msg = "Improper type for `rg._src`."
            raise ValueError(e_msg)
        _add_mask_to_cube(rg.src_mask, src_cube, _SOURCE_MASK_NAME)

        tgt_grid = rg._tgt
        if isinstance(tgt_grid, GridRecord):
            tgt_cube = _standard_grid_cube(
                (tgt_grid.grid_y, tgt_grid.grid_x), _TARGET_NAME
            )
        elif isinstance(tgt_grid, MeshRecord):
            tgt_mesh, tgt_location = tgt_grid
            tgt_cube = _standard_mesh_cube(tgt_mesh, tgt_location, _TARGET_NAME)
        else:
            e_msg = "Improper type for `rg._tgt`."
            raise ValueError(e_msg)
        _add_mask_to_cube(rg.tgt_mask, tgt_cube, _TARGET_MASK_NAME)
    elif regridder_type == "GridToMeshESMFRegridder":
        src_grid = (rg.grid_y, rg.grid_x)
        src_cube = _standard_grid_cube(src_grid, _SOURCE_NAME)
        _add_mask_to_cube(rg.src_mask, src_cube, _SOURCE_MASK_NAME)

        tgt_mesh = rg.mesh
        tgt_location = rg.location
        tgt_cube = _standard_mesh_cube(tgt_mesh, tgt_location, _TARGET_NAME)
        _add_mask_to_cube(rg.tgt_mask, tgt_cube, _TARGET_MASK_NAME)

    elif regridder_type == "MeshToGridESMFRegridder":
        src_mesh = rg.mesh
        src_location = rg.location
        src_cube = _standard_mesh_cube(src_mesh, src_location, _SOURCE_NAME)
        _add_mask_to_cube(rg.src_mask, src_cube, _SOURCE_MASK_NAME)

        tgt_grid = (rg.grid_y, rg.grid_x)
        tgt_cube = _standard_grid_cube(tgt_grid, _TARGET_NAME)
        _add_mask_to_cube(rg.tgt_mask, tgt_cube, _TARGET_MASK_NAME)

    else:
        e_msg = f"Unexpected regridder type {regridder_type}."
        raise TypeError(e_msg)
    return src_cube, tgt_cube


def save_regridder(rg, filename, allow_partial=False):
    """Save a regridder scheme instance.

    Saves any of the regridder classes, i.e.
    :class:`~esmf_regrid.experimental.unstructured_scheme.GridToMeshESMFRegridder`,
    :class:`~esmf_regrid.experimental.unstructured_scheme.MeshToGridESMFRegridder`,
    :class:`~esmf_regrid.schemes.ESMFAreaWeightedRegridder`,
    :class:`~esmf_regrid.schemes.ESMFBilinearRegridder` or
    :class:`~esmf_regrid.schemes.ESMFNearestRegridder`.
    .

    Parameters
    ----------
    rg : :class:`~esmf_regrid.schemes._ESMFRegridder`
        The regridder instance to save.
    filename : str
        The file name to save to.
    allow_partial : bool, default=False
        If True, allow the saving of :class:`~esmf_regrid.experimental._partial.PartialRegridder` instances.
    """
    regridder_type = rg.__class__.__name__

    src_cube, tgt_cube = _generate_src_tgt(regridder_type, rg, allow_partial)

    method = str(check_method(rg.method).name)

    if regridder_type in ["GridToMeshESMFRegridder", "MeshToGridESMFRegridder"]:
        resolution = rg.resolution
        src_resolution = None
        tgt_resolution = None
    elif method == "CONSERVATIVE":
        resolution = None
        src_resolution = rg.src_resolution
        tgt_resolution = rg.tgt_resolution
    else:
        resolution = None
        src_resolution = None
        tgt_resolution = None

    weight_matrix = rg.regridder.weight_matrix
    reformatted_weight_matrix = scipy.sparse.coo_matrix(weight_matrix)
    weight_data = reformatted_weight_matrix.data
    weight_rows = reformatted_weight_matrix.row
    weight_cols = reformatted_weight_matrix.col
    weight_shape = reformatted_weight_matrix.shape

    esmf_version = rg.regridder.esmf_version
    esmf_regrid_version = rg.regridder.esmf_regrid_version
    save_version = esmf_regrid.__version__

    # Currently, all schemes use the fracarea normalization.
    normalization = "fracarea"

    mdtol = rg.mdtol
    attributes = {
        "title": "iris-esmf-regrid regridding scheme",
        _REGRIDDER_TYPE: regridder_type,
        _VERSION_ESMF: esmf_version,
        _VERSION_INITIAL: esmf_regrid_version,
        "esmf_regrid_version_on_save": save_version,
        "normalization": normalization,
        _MDTOL: mdtol,
        _METHOD: method,
    }
    if resolution is not None:
        attributes[_RESOLUTION] = resolution
    if src_resolution is not None:
        attributes[_SOURCE_RESOLUTION] = src_resolution
    if tgt_resolution is not None:
        attributes[_TARGET_RESOLUTION] = tgt_resolution

    extra_cubes = []
    if regridder_type == "PartialRegridder":
        src_slice = rg.src_slice  # this slice is described by a tuple
        if src_slice is None:
            src_slice = []
        src_slice_cube = Cube(
            src_slice, long_name=_SRC_SLICE_NAME, var_name=_SRC_SLICE_NAME
        )
        tgt_slice = rg.tgt_slice  # this slice is described by a tuple
        if tgt_slice is None:
            tgt_slice = []
        tgt_slice_cube = Cube(
            tgt_slice, long_name=_TGT_SLICE_NAME, var_name=_TGT_SLICE_NAME
        )
        extra_cubes = [src_slice_cube, tgt_slice_cube]

    weights_cube = Cube(weight_data, var_name=_WEIGHTS_NAME, long_name=_WEIGHTS_NAME)
    row_coord = AuxCoord(
        weight_rows, var_name=_WEIGHTS_ROW_NAME, long_name=_WEIGHTS_ROW_NAME
    )
    col_coord = AuxCoord(
        weight_cols, var_name=_WEIGHTS_COL_NAME, long_name=_WEIGHTS_COL_NAME
    )
    weights_cube.add_aux_coord(row_coord, 0)
    weights_cube.add_aux_coord(col_coord, 0)

    esmf_args = rg.esmf_args
    if esmf_args is None:
        esmf_args = {}
    for arg in esmf_args:
        if arg not in _VALID_ESMF_KWARGS:
            e_msg = f"{arg} is not considered a valid argument to pass to ESMF."
            raise KeyError(e_msg)
    esmf_arg_attributes = {
        k: v.name if hasattr(v, "name") else int(v) if isinstance(v, bool) else v
        for k, v in esmf_args.items()
    }
    esmf_arg_coord = AuxCoord(
        0, var_name=_ESMF_ARGS, long_name=_ESMF_ARGS, attributes=esmf_arg_attributes
    )
    weights_cube.add_aux_coord(esmf_arg_coord)

    weight_shape_cube = Cube(
        weight_shape,
        var_name=_WEIGHTS_SHAPE_NAME,
        long_name=_WEIGHTS_SHAPE_NAME,
    )

    # Save cubes while ensuring var_names do not conflict for the sake of consistency.
    with _managed_var_name(src_cube, tgt_cube):
        cube_list = CubeList(
            [src_cube, tgt_cube, weights_cube, weight_shape_cube, *extra_cubes]
        )

        for cube in cube_list:
            cube.attributes = attributes

        iris.fileformats.netcdf.save(cube_list, filename)


def load_regridder(filename, allow_partial=False):
    """Load a regridder scheme instance.

    Loads any of the regridder classes, i.e.
    :class:`~esmf_regrid.experimental.unstructured_scheme.GridToMeshESMFRegridder`,
    :class:`~esmf_regrid.experimental.unstructured_scheme.MeshToGridESMFRegridder`,
    :class:`~esmf_regrid.schemes.ESMFAreaWeightedRegridder`,
    :class:`~esmf_regrid.schemes.ESMFBilinearRegridder` or
    :class:`~esmf_regrid.schemes.ESMFNearestRegridder`.

    Parameters
    ----------
    filename : str
        The file name to load from.
    allow_partial : bool, default=False
        If True, allow the loading of :class:`~esmf_regrid.experimental._partial.PartialRegridder` instances.

    Returns
    -------
    :class:`~esmf_regrid.schemes._ESMFRegridder`
    """
    with _load_context():
        cubes = iris.load(filename)

    # Extract the source, target and metadata information.
    src_cube = cubes.extract_cube(_SOURCE_NAME)
    _clean_var_names(src_cube)
    tgt_cube = cubes.extract_cube(_TARGET_NAME)
    _clean_var_names(tgt_cube)
    weights_cube = cubes.extract_cube(_WEIGHTS_NAME)
    weight_shape_cube = cubes.extract_cube(_WEIGHTS_SHAPE_NAME)

    # Determine the regridder type.
    regridder_type = weights_cube.attributes[_REGRIDDER_TYPE]
    if regridder_type not in _REGRIDDER_NAME_MAP:
        e_msg = f"Unrecognised regridder type {regridder_type}."
        raise TypeError(e_msg)
    scheme = _REGRIDDER_NAME_MAP[regridder_type]

    if regridder_type == "PartialRegridder" and not allow_partial:
        e_msg = (
            "PartialRegridder cannot be loaded without setting `allow_partial=True`."
        )
        raise ValueError(e_msg)

    # Determine the regridding method, allowing for files created when
    # conservative regridding was the only method.
    method_string = weights_cube.attributes.get(_METHOD, "CONSERVATIVE")
    # Account for strings saved in previous versions.
    method_string = method_string.upper()
    method = getattr(Constants.Method, method_string)

    resolution = weights_cube.attributes.get(_RESOLUTION, None)
    src_resolution = weights_cube.attributes.get(_SOURCE_RESOLUTION, None)
    tgt_resolution = weights_cube.attributes.get(_TARGET_RESOLUTION, None)
    if resolution is not None:
        resolution = int(resolution)
    if src_resolution is not None:
        src_resolution = int(src_resolution)
    if tgt_resolution is not None:
        tgt_resolution = int(tgt_resolution)

    # Reconstruct the weight matrix.
    weight_data = weights_cube.data
    weight_rows = weights_cube.coord(_WEIGHTS_ROW_NAME).points
    weight_cols = weights_cube.coord(_WEIGHTS_COL_NAME).points
    weight_shape = weight_shape_cube.data
    weight_matrix = scipy.sparse.csr_matrix(
        (weight_data, (weight_rows, weight_cols)), shape=weight_shape
    )

    mdtol = weights_cube.attributes[_MDTOL]

    if src_cube.coords(_SOURCE_MASK_NAME):
        use_src_mask = src_cube.coord(_SOURCE_MASK_NAME).points.astype(bool)
    else:
        use_src_mask = False
    if tgt_cube.coords(_TARGET_MASK_NAME):
        use_tgt_mask = tgt_cube.coord(_TARGET_MASK_NAME).points.astype(bool)
    else:
        use_tgt_mask = False

    # Allow for this coord not to exist for the sake of backwards compatibility.
    esmf_args_coords = weights_cube.coords(_ESMF_ARGS)
    if len(esmf_args_coords) == 0:
        esmf_args = {}
    else:
        esmf_args = esmf_args_coords[0].attributes
    for arg, arg_dict in _ESMF_ENUM_ARGS.items():
        if arg in esmf_args:
            esmf_args[arg] = arg_dict[esmf_args[arg]]
    for arg in _ESMF_BOOL_ARGS:
        if arg in esmf_args:
            esmf_args[arg] = bool(esmf_args[arg])

    if scheme is GridToMeshESMFRegridder:
        resolution_keyword = _SOURCE_RESOLUTION
        kwargs = {resolution_keyword: resolution, "method": method, "mdtol": mdtol}
    elif scheme is MeshToGridESMFRegridder:
        resolution_keyword = _TARGET_RESOLUTION
        kwargs = {resolution_keyword: resolution, "method": method, "mdtol": mdtol}
    elif method is Constants.Method.CONSERVATIVE:
        kwargs = {
            _SOURCE_RESOLUTION: src_resolution,
            _TARGET_RESOLUTION: tgt_resolution,
            "mdtol": mdtol,
        }
    elif method is Constants.Method.BILINEAR:
        kwargs = {"mdtol": mdtol}
    else:
        kwargs = {}

    if scheme is PartialRegridder:
        src_slice = cubes.extract_cube(_SRC_SLICE_NAME).data.tolist()
        if src_slice == []:
            src_slice = None
        tgt_slice = cubes.extract_cube(_TGT_SLICE_NAME).data.tolist()
        if tgt_slice == []:
            tgt_slice = None
        sub_scheme = {
            Constants.Method.CONSERVATIVE: ESMFAreaWeighted,
            Constants.Method.BILINEAR: ESMFBilinear,
        }[method]
        mdtol = kwargs.pop(_MDTOL, None)
        sub_kwargs = {}
        if mdtol is not None:
            sub_kwargs[_MDTOL] = mdtol
        regridder = scheme(
            src_cube,
            tgt_cube,
            src_slice,
            tgt_slice,
            weight_matrix,
            sub_scheme(
                use_src_mask=use_src_mask,
                use_tgt_mask=use_tgt_mask,
                esmf_args=esmf_args,
                **sub_kwargs,
            ),
            **kwargs,
        )
    else:
        regridder = scheme(
            src_cube,
            tgt_cube,
            precomputed_weights=weight_matrix,
            use_src_mask=use_src_mask,
            use_tgt_mask=use_tgt_mask,
            esmf_args=esmf_args,
            **kwargs,
        )

    esmf_version = weights_cube.attributes[_VERSION_ESMF]
    regridder.regridder.esmf_version = esmf_version
    esmf_regrid_version = weights_cube.attributes[_VERSION_INITIAL]
    regridder.regridder.esmf_regrid_version = esmf_regrid_version
    return regridder



"""Unit tests for :mod:`esmf_regrid.experimental.io.load_regridder`."""

import numpy as np
from numpy import ma

from esmf_regrid.experimental.io import load_regridder
from esmf_regrid.experimental.unstructured_scheme import GridToMeshESMFRegridder
from esmf_regrid.tests.unit.experimental.io.test_round_tripping import (
    _compare_ignoring_var_names,
    _make_grid_to_mesh_regridder,
)


def test_load_v09(request):
    """Test the loading of a file saved in version 0.9.0."""
    filename = request.path.parent / "v0.9_regridder.nc"
    method = "conservative"
    regridder = GridToMeshESMFRegridder
    equivalent_rg, src = _make_grid_to_mesh_regridder(
        method=method, regridder=regridder, circular=True
    )
    loaded_rg = load_regridder(str(filename))

    assert equivalent_rg.location == loaded_rg.location
    assert equivalent_rg.method == loaded_rg.method
    assert equivalent_rg.mdtol == loaded_rg.mdtol
    _compare_ignoring_var_names(equivalent_rg.grid_x, loaded_rg.grid_x)
    _compare_ignoring_var_names(equivalent_rg.grid_y, loaded_rg.grid_y)

    # Compare the weight matrices.
    original_matrix = equivalent_rg.regridder.weight_matrix
    loaded_matrix = loaded_rg.regridder.weight_matrix
    # Ensure the original and loaded weight matrix have identical type.
    assert type(original_matrix) is type(loaded_matrix)  # E721
    assert np.array_equal(original_matrix.todense(), loaded_matrix.todense())

    # Demonstrate regridding still gives the same results.
    src_data = ma.arange(np.prod(src.data.shape)).reshape(src.data.shape)
    src_data[0, 0] = ma.masked
    src.data = src_data
    # TODO: make this a cube comparison when mesh comparison becomes available.
    original_result = equivalent_rg(src).data
    loaded_result = loaded_rg(src).data
    assert np.array_equal(original_result, loaded_result)
    assert np.array_equal(original_result.mask, loaded_result.mask)

    # Ensure versions match what was saved.
    expected_esmf_version = "8.4.2"
    expected_save_version = "0.9.0"
    assert expected_esmf_version == loaded_rg.regridder.esmf_version
    assert expected_save_version == loaded_rg.regridder.esmf_regrid_version



"""Unit tests for :mod:`esmf_regrid.experimental.partition`."""

import numpy as np

from esmf_regrid import ESMFAreaWeighted
from esmf_regrid.experimental._partial import PartialRegridder
from esmf_regrid.experimental.io import load_regridder, save_regridder
from esmf_regrid.tests.unit.schemes.test__cube_to_GridInfo import (
    _grid_cube,
)


def test_PartialRegridder_repr():
    """Test repr of PartialRegridder instance."""
    src = _grid_cube(10, 15, (-180, 180), (-90, 90), circular=True)
    tgt = _grid_cube(5, 10, (-180, 180), (-90, 90), circular=True)
    src_slice = ((10, 20), (15, 30))
    tgt_slice = ((0, 5), (0, 10))
    weights = None
    scheme = ESMFAreaWeighted(mdtol=0.5)

    pr = PartialRegridder(src, tgt, src_slice, tgt_slice, weights, scheme)

    expected_repr = (
        "PartialRegridder(src=GridRecord("
        "grid_x=<DimCoord: longitude / (degrees)  [-162., -126., ..., 126., 162.]+bounds  shape(10,)>, "
        "grid_y=<DimCoord: latitude / (degrees)  [-84., -72., ..., 72., 84.]+bounds  shape(15,)>), "
        "tgt_slice=GridRecord(grid_x=<DimCoord: longitude / (degrees)  [-144., -72., 0., 72., 144.]+bounds  shape(5,)>, "
        "grid_y=<DimCoord: latitude / (degrees)  [-81., -63., ..., 63., 81.]+bounds  shape(10,)>), "
        "src_slice=((10, 20), (15, 30)), tgt_slice=((0, 5), (0, 10)), scheme=ESMFAreaWeighted(mdtol=0.5, "
        "use_src_mask=False, use_tgt_mask=False, esmf_args={}))"
    )
    assert repr(pr) == expected_repr


def test_PartialRegridder_roundtrip(tmp_path):
    """Test load/save for PartialRegridder instance."""
    src = _grid_cube(10, 15, (-180, 180), (-90, 90), circular=True)
    mask = np.zeros_like(src.data)
    mask[0, 0] = 1
    src.data = np.ma.array(src.data, mask=mask)
    tgt = _grid_cube(5, 10, (-180, 180), (-90, 90), circular=True)
    src_slice = [[10, 20], [15, 30]]
    tgt_slice = [[0, 5], [0, 10]]
    weights = None
    scheme = ESMFAreaWeighted(
        mdtol=0.5, use_src_mask=src.data.mask, esmf_args={"ignore_degenerate": True}
    )

    pr = PartialRegridder(src, tgt, src_slice, tgt_slice, weights, scheme)
    file = tmp_path / "partial.nc"

    save_regridder(pr, file, allow_partial=True)
    loaded_pr = load_regridder(file, allow_partial=True)

    assert repr(loaded_pr) == repr(pr)



"""Provides a regridder class compatible with Partition."""

import numpy as np

from esmf_regrid.schemes import (
    GridRecord,
    MeshRecord,
    _create_cube,
    _ESMFRegridder,
)


class PartialRegridder(_ESMFRegridder):
    """Regridder class designed for use in :class:`~esmf_regrid.experimental.Partition`."""

    def __init__(self, src, tgt, src_slice, tgt_slice, weights, scheme, **kwargs):
        """Create a regridder instance for a block of :class:`~esmf_regrid.experimental.Partition`.

        Parameters
        ----------
        src : :class:`iris.cube.Cube`
            The :class:`~iris.cube.Cube` providing the source.
        tgt : :class:`iris.cube.Cube` or :class:`iris.mesh.MeshXY`
            The :class:`~iris.cube.Cube` or :class:`~iris.mesh.MeshXY` providing the target.
        src_slice : tuple
            The upper and lower bounds of the block taken from the original source from which the
            ``src`` was derived. In the form ((x_low, x_high), ...) where x_low and x_high are the
            upper and lower bounds of the slice (in the x dimension) taken from the original source.
            There are as many tuples of upper and lower bounds as there are horizontal dimensions in
            the source cube (currently this is always 2 as Meshes are not yet supported for sources).
        tgt_slice : tuple
            The upper and lower bounds of the block taken from the original target from which the
            ``tgt`` was derived. In the form ((x_low, x_high), ...) where x_low and x_high are the
            upper and lower bounds of the slice (in the x dimension) taken from the original target.
            There are as many tuples of upper and lower bounds as there are horizontal dimensions in
            the target cube.
        weights : :class:`scipy.sparse.spmatrix`
            The weights to use for regridding.
        scheme : :class:`~esmf_regrid.schemes.ESMFAreaWeighted` or :class:`~esmf_regrid.schemes.ESMFBilinear`
            The scheme used to construct the regridder.
        kwargs : dict
            Additional keyword arguments to pass to the `scheme`s regridder method.
        """
        self.src_slice = src_slice  # this will be tuple-like
        self.tgt_slice = tgt_slice
        self.scheme = scheme

        self._regridder = scheme.regridder(
            src,
            tgt,
            precomputed_weights=weights,
            **kwargs,
        )
        self.__dict__.update(self._regridder.__dict__)

    def __repr__(self):
        """Return a representation of the class."""
        result = (
            f"PartialRegridder("
            f"src={self._src}, "
            f"tgt_slice={self._tgt}, "
            f"src_slice={self.src_slice}, "
            f"tgt_slice={self.tgt_slice}, "
            f"scheme={self.scheme})"
        )
        return result

    def partial_regrid(self, src):
        """Perform the first half of regridding, generating weights and data."""
        dims = self._get_cube_dims(src)
        num_dims = len(dims)
        standard_in_dims = [-1, -2][:num_dims]
        data = np.moveaxis(src.data, dims, standard_in_dims)
        result = self.regridder._gen_weights_and_data(data)
        return result

    def finish_regridding(self, src_cube, weights, data, extra):
        """Perform the second half of regridding, combining weights and data.

        This operation is used to process the combined results from all the partial
        regridders in a Partition.
        Since all the combined data is passed in, this operation can be done using
        *any one* of the individual PartialRegridders.
        However, the passed "src_cube" must be the "correct" slice of the source
        data cube, corresponding to the 'tgt_slice' slice params it was created with.
        It is also implicit that the 'extra' arg (additional dimensions) will be the
        same for all partial results.
        The `src_cube` provides coordinates for the non-horizontal dimensions of the
        result cube, matching the dimensions of the `data` array.
        For technical convenience, its *horizontal* coordinates need to match those
        of the 'src' reference cube provided in regridder creation (`self._src`).
        So, it must be the correct "corresponding slice" of the source cube.
        """
        src_dims = self._get_cube_dims(src_cube)

        result_data = self.regridder._regrid_from_weights_and_data(weights, data, extra)

        num_out_dims = self.regridder.tgt.dims
        num_dims = len(src_dims)
        standard_out_dims = [-1, -2][:num_out_dims]
        if num_dims == 2 and num_out_dims == 1:
            new_dims = [min(src_dims)]
        elif num_dims == 1 and num_out_dims == 2:
            # Note: this code is currently inaccessible since src_cube can't have a Mesh.
            new_dims = [src_dims[0] + 1, src_dims[0]]
        else:
            new_dims = src_dims

        result_data = np.moveaxis(result_data, standard_out_dims, new_dims)

        if isinstance(self._tgt, GridRecord):
            tgt_coords = self._tgt
            out_dims = 2
        elif isinstance(self._tgt, MeshRecord):
            tgt_coords = self._tgt.mesh.to_MeshCoords(self._tgt.location)
            out_dims = 1
        else:
            msg = "Unrecognised target information."
            raise TypeError(msg)

        result_cube = _create_cube(
            result_data, src_cube, src_dims, tgt_coords, out_dims
        )
        return result_cube



"""Perform test automation with nox.

For further details, see https://nox.thea.codes/en/stable/#

"""

import os
from pathlib import Path
import shutil
from urllib.error import HTTPError
from urllib.parse import urlparse
from urllib.request import urlopen

import nox
from nox.logger import logger
import yaml

#: Default to reusing any pre-existing nox environments.
nox.options.reuse_existing_virtualenvs = False

#: Name of the package to test.
PACKAGE = "esmf_regrid"

#: Python versions we can run sessions under
_PY_VERSIONS_ALL = ["3.12", "3.13", "3.14"]

#: GHA-CI environment variable hook.
PY_VER = os.environ.get("PY_VER", _PY_VERSIONS_ALL)

#: GHA-CI environment variable hook.
COVERAGE = os.environ.get("COVERAGE", "")

#: GHA-CI environment variable hook.
#: If you change the IRIS_SOURCE here you will also need to change it in
#: the tests, wheel and benchmark workflows.
IRIS_SOURCE = os.environ.get("IRIS_SOURCE", "github:main")

IRIS_GITHUB = "https://github.com/scitools/iris.git"
LOCKFILE_PLATFORM = "linux-64"

ESMFMKFILE = "ESMFMKFILE"


def _lockfile_path(py_string: str, platform_placeholder: bool = False) -> Path:
    """Return a constructed lockfile path for the relevant python string e.g ``py38``.

    Optionally retain the ``{platform}`` placeholder to support conda-lock's
    ``--filename-template``.

    """
    lockfile_dir = Path() / "requirements" / "locks"
    name_template = "{py_string}-{platform}.lock"
    if platform_placeholder:
        platform = "{platform}"
    else:
        platform = LOCKFILE_PLATFORM
    lockfile_name = name_template.format(py_string=py_string, platform=platform)
    return lockfile_dir / lockfile_name


def _session_lockfile(session: nox.sessions.Session) -> Path:
    """Return the path of the session lockfile."""
    return _lockfile_path(py_string=f"py{session.python.replace('.', '')}")


def _file_content(file_path: Path) -> str:
    with file_path.open("r") as file:
        return file.read()


def _session_cachefile(session: nox.sessions.Session) -> Path:
    """Return the path of the session lockfile cache."""
    tmp_dir = Path(session.create_tmp())
    cache = tmp_dir / _session_lockfile(session).name
    return cache


def _venv_populated(session: nox.sessions.Session) -> bool:
    """Return True if the Conda venv has been created and the list of packages in the lockfile installed."""
    return _session_cachefile(session).is_file()


def _venv_changed(session: nox.sessions.Session) -> bool:
    """Return True if the installed session is different to that specified in the lockfile."""
    result = False
    if _venv_populated(session):
        expected = _file_content(_session_lockfile(session))
        actual = _file_content(_session_cachefile(session))
        result = actual != expected
    return result


def _install_and_cache_venv(session: nox.sessions.Session) -> None:
    """Cache the nox session environment.

    This consists of saving a hexdigest (sha256) of the associated
    Conda lock file.

    Parameters
    ----------
    session : object
        A `nox.sessions.Session` object.

    """
    lockfile = _session_lockfile(session)
    session.conda_install(f"--file={lockfile}")
    with _session_cachefile(session).open("w") as cachefile:
        cachefile.write(_file_content(lockfile))


def _get_iris_github_artifact(session: nox.sessions.Session) -> str:
    """Determine whether an Iris source artifact from GitHub is required.

    This can be an Iris branch name, commit sha or tag name.

    Parameters
    ----------
    session : object
        A `nox.sessions.Session` object.

    Returns
    -------
    str
        The Iris GitHub artifact.

    """
    result = IRIS_SOURCE

    # The CLI overrides the environment variable.
    for arg in session.posargs:
        if arg.startswith("--iris="):
            parts = arg.split("=")
            if len(parts) == 2:
                result = parts[1].strip()
                break
    if result:
        parts = result.strip().split(":")
        result = None
        if len(parts) == 2:
            repo, artifact = parts
            if repo.startswith(("'", '"')):
                repo = repo[1:]
            if repo.lower() == "github":
                result = artifact
                if result.endswith(("'", '"')):
                    result = result[:-1]
    return result


def _prepare_env(session: nox.sessions.Session) -> None:
    venv_dir = session.virtualenv.location_name

    esmf_mk_file = Path(venv_dir) / "lib" / "esmf.mk"
    session.env[ESMFMKFILE] = esmf_mk_file.absolute()

    if not _venv_populated(session):
        # Environment has been created but packages not yet installed.
        # Populate the environment from the lockfile.
        logger.debug(f"Populating conda env: {venv_dir}")
        _install_and_cache_venv(session)

    elif _venv_changed(session):
        # Destroy the environment and rebuild it.
        logger.debug(f"Lockfile changed. Recreating conda env: {venv_dir}")
        _reuse_original = session.virtualenv.reuse_existing
        session.virtualenv.reuse_existing = False
        session.virtualenv.create()
        _install_and_cache_venv(session)
        session.virtualenv.reuse_existing = _reuse_original

    logger.debug(f"Environment up to date: {venv_dir}")

    iris_artifact = _get_iris_github_artifact(session)
    if iris_artifact:
        # Install the iris source in develop mode.
        tmp_dir = Path(session.create_tmp())
        iris_dir = tmp_dir / "iris"
        cwd = Path.cwd()
        if not iris_dir.is_dir():
            session.run_always(
                "git", "clone", IRIS_GITHUB, str(iris_dir), external=True
            )
        session.cd(str(iris_dir))
        session.run_always("git", "fetch", "origin", external=True)
        session.run_always("git", "checkout", iris_artifact, external=True)
        session.cd(str(cwd))
        session.install("--no-deps", "--editable", str(iris_dir))

    # Determine whether verbose diagnostics have been requested
    # from the command line.
    verbose = "-v" in session.posargs or "--verbose" in session.posargs

    if verbose:
        session.run_always("conda", "info")
        session.run_always("conda", "list", f"--prefix={venv_dir}")
        session.run_always(
            "conda",
            "list",
            f"--prefix={venv_dir}",
            "--explicit",
        )


@nox.session
def update_lockfiles(session: nox.sessions.Session):
    """Re-resolve env specs and store as lockfiles (``requirements/locks/``).

    Original Conda environment specifications are at:
    ``requirements/py**.yml``. The output lock files denote the dependencies
    that iris-esmf-regrid is tested against, and therefore officially supports.

    Parameters
    ----------
    session : object
        A `nox.sessions.Session` object.

    """
    session.install("conda-lock")

    for req_file in Path("requirements").glob(r"py[0-9]*.yml"):
        python_string = req_file.stem

        # Generate the appropriate conda-lock template name, keeping the {platform}
        # placeholder to support conda-lock's internals.
        filename_template = _lockfile_path(python_string, platform_placeholder=True)
        lockfile_path = _lockfile_path(python_string, platform_placeholder=False)
        # Create the parent directory if it doesn't already exist.
        try:
            filename_template.parent.mkdir()
        except FileExistsError:
            pass

        # Use a copy of the requirements file in a tmp dir - the file will
        # be modified if installing a custom Iris checkout.
        tmp_dir = Path(session.create_tmp())
        req_file_local = tmp_dir / req_file.name
        shutil.copy(req_file, req_file_local)

        conda_lock_cmd = [
            "conda-lock",
            "lock",
            "--kind=explicit",
            f"--filename-template={filename_template}",
            f"--file={req_file_local}",
            f"--platform={LOCKFILE_PLATFORM}",
        ]

        # Get the requirements for Iris too, if an Iris checkout is specified.
        iris_artifact = _get_iris_github_artifact(session)
        if iris_artifact:
            # Remove ``iris`` from dependencies, if present.
            with req_file_local.open("r+") as file:
                reqs = yaml.load(file, Loader=yaml.FullLoader)
                reqs["dependencies"] = [
                    spec for spec in reqs["dependencies"] if not spec.startswith("iris")
                ]
                yaml.dump(reqs, file)

            iris_req_name = f"{python_string}.yml"
            iris_req_url = (
                f"https://raw.githubusercontent.com/SciTools/iris/"
                f"{iris_artifact}/requirements/ci/{iris_req_name}"
            )
            iris_req_file = (tmp_dir / iris_req_name).with_stem(f"{python_string}-iris")

            try:
                # first attempt with legacy requirements structure
                connection = urlopen(iris_req_url)
            except HTTPError as error:
                if error.code == 404:
                    # retry with new requirements structure i.e., no "ci" directory
                    url = urlparse(iris_req_url)
                    parts = url.path.split("/")
                    parts.remove("ci")
                    url = url._replace(path="/".join(parts))
                    connection = urlopen(url.geturl())

            iris_req = connection.read()
            with iris_req_file.open("wb") as fout:
                fout.write(iris_req)
            # Conda-lock can resolve multiple requirements files together.
            conda_lock_cmd.append(f"--file={iris_req_file}")

        session.run(*conda_lock_cmd, silent=True)
        print(f"Conda lock file created: {lockfile_path}")


@nox.session(python=PY_VER, venv_backend="conda")
def tests(session: nox.sessions.Session):
    """Perform esmf-regrid integration and unit tests.

    Parameters
    ----------
    session : object
        A `nox.sessions.Session` object.

    """
    _prepare_env(session)
    # Install the esmf-regrid source in develop mode.
    session.install("--no-deps", "--editable", ".")

    if COVERAGE:
        # Execute the tests with code coverage.
        session.run("pytest", "--cov-report=xml", "--cov")
    else:
        # Execute the tests.
        session.run("pytest")


@nox.session(python=PY_VER, venv_backend="conda")
def wheel(session: nox.sessions.Session):
    """Perform iris-esmf-regrid local wheel install and import test.

    Parameters
    ----------
    session : object
        A `nox.sessions.Session` object.

    """
    _prepare_env(session)
    session.cd("dist")
    fname = list(Path().glob("esmf_regrid-*.whl"))
    if len(fname) == 0:
        e_msg = "Cannot find wheel to install."
        raise ValueError(e_msg)
    if len(fname) > 1:
        emsg = f"Expected to find 1 wheel to install, found {len(fname)} instead."
        raise ValueError(emsg)
    session.install(fname[0].name)
    session.run(
        "python",
        "-c",
        "import esmf_regrid; print(f'{esmf_regrid.__version__=}')",
        external=True,
    )
